#include "FxCyanRGB8.hpp"

// C++ system level
#include <string>
#include <cstring>  // memset
#include <cinttypes>
#include <algorithm>
#include <functional>

// ESP32 specific
// #include "freertos/FreeRTOS.h"
// #include "freertos/queue.h"

// project specific
#include "FxVSync.hpp"
#include "CyanBus.hpp"

// qthing stuff
// #include ""



namespace SiliconTorch {

  namespace FxCyanRGB8 {

    const std::string HEADER("fxCyanRGB8");


    FxCyanRGB8::FxCyanRGB8(uint16_t leds, uint16_t startIdx) : leds(leds), startIdx(startIdx) {

      buffer0 = new qthing::RGB[leds];
      buffer1 = new qthing::RGB[leds];

      // std::memset(buffer0, 0x00, leds * sizeof(Pixel));
      // std::memset(buffer1, 0x00, leds * sizeof(Pixel));



    }

    void FxCyanRGB8::handlePacket(uint8_t* buffer, std::size_t length) {

      if (length < 3) {
        ESP_LOGE(HEADER.c_str(), "Wrong |buffer|[ %d ]: Must be greater than 2!", length);
        return;
      }

      if ((length - 2) % 3 > 0) {
        ESP_LOGE(HEADER.c_str(), "Wrong |buffer|[ %d ]: (Length - 2) must be divisible by 3!", length);
        return;
      }


      Pixel* packetPx = reinterpret_cast<Pixel*>(buffer+2);
      uint16_t packetPxs = (length - 2) / 3;
      uint16_t packetStartIdx = (buffer[0] << 8) | buffer[1];


      if (packetStartIdx == startIdx) {

        for (uint16_t px = 0; px < std::min(leds, packetPxs); px++) {
          buffer1[px].r = packetPx[px].r / 255.0f;
          buffer1[px].g = packetPx[px].g / 255.0f;
          buffer1[px].b = packetPx[px].b / 255.0f;
        }

      } else if(packetStartIdx > startIdx) {

        uint16_t delta = packetStartIdx - startIdx;

        for (uint16_t px = delta; px < std::min(leds, (uint16_t)(packetPxs + delta)); px++) {
          buffer1[px].r = packetPx[px - delta].r / 255.0f;
          buffer1[px].g = packetPx[px - delta].g / 255.0f;
          buffer1[px].b = packetPx[px - delta].b / 255.0f;
        }

      } else if (packetStartIdx < startIdx) {

        uint16_t delta = startIdx - packetStartIdx;

        // Packet does not contain any pixels for us
        if (packetPxs < delta) return;

        for (uint16_t px = 0; px < std::min(leds, (uint16_t)(packetPxs - delta)); px++) {
          buffer1[px].r = packetPx[px + delta].r / 255.0f;
          buffer1[px].g = packetPx[px + delta].g / 255.0f;
          buffer1[px].b = packetPx[px + delta].b / 255.0f;
        }
      }
    }

    void FxCyanRGB8::registerAtCyanBus(CyanBus::CyanBus& cyanBus) {

      uint8_t headerLength = HEADER.length();

      cyanBus.packetCallback.add(HEADER, [&, headerLength](const SiliconTorch::CyanBus::PacketData& packet) {
        this->handlePacket((uint8_t*)(packet.payload + headerLength), packet.length - headerLength);
      });

      FxVSync::registerAtCyanBus(cyanBus, [&]() {
        requestBufferSwap();
      });

    }

    void FxCyanRGB8::registerUDPHandler() {
      if (!udpHandlerRegistered) {
        udpHandlerRegistered = true;

        uint8_t headerLength = HEADER.length();

        qthing::addUDPPacketCallback(HEADER, [&, headerLength](qthing::udpPacket packet) {
          this->handlePacket((uint8_t*)(packet.payload + headerLength), packet.length - headerLength);
          requestBufferSwap();
        });
      }
    }

    // qthing::Animator
    qthing::RGB FxCyanRGB8::render(uint16_t addr) {
      if (addr < leds) {
        return buffer0[addr];
      } else {
        qthing::RGB black;
        return black;
      }
    }

    // qthing::Animator
    void FxCyanRGB8::step() {
      if (swapRequested) swapBuffers();
    }

    void FxCyanRGB8::requestBufferSwap() {
      swapRequested = true;
    }

    void FxCyanRGB8::swapBuffers() {
      swapRequested = false;
      qthing::RGB* tmp = buffer0;
      buffer0 = buffer1;
      buffer1 = tmp;

      for (uint16_t px = 0; px < leds; px++){
        buffer1[px].r = 0;
        buffer1[px].g = 0;
        buffer1[px].b = 0;
      }
    }
  }
}
