#pragma once

// C++ system level
// #include <cstring>  // memset, strncmp
// #include <cstdlib>  // TODO: is this for memcpy?
// #include <cinttypes>
#include <functional>

// ESP32 specific
// #include "esp_log.h"

// project specific
// #include "CyanBusCRC.hpp"

// qthing stuff
// #include ""


namespace SpiderLib {

  // Forward declaration
  template <class>
  class Callback;


  template <typename... Args>
  class Callback<void(Args...)> {

    public:
      typedef std::function<void(Args...)> CallbackF;

      Callback() : callback(NULL) {};

      void operator()(Args&&... args) const {
        if (callback != NULL) callback(std::forward<Args>(args)...);
      }

      auto& operator+=(const CallbackF& newCB) const {
        add(newCB);
        return *this;
      }

      void add(CallbackF newCB) const {
        if (callback == NULL) {
          callback = newCB;
        } else {
          auto oldCB = callback;
          callback = [&](Args&&... args) {
            oldCB(std::forward<Args>(args)...);
            newCB(std::forward<Args>(args)...);
          };
        }
      }

    private:
      mutable CallbackF callback = NULL;
  };

}
